{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# CASE WHEN behavior change"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**About the change**\n",
    "\n",
    "The behavior of the CASE WHEN operator in PQL will be changed in regards to the behavior when a condition evaluates to NULL. In the future, a condition that evaluates to NULL is always treated as if it was false. The new behavior is aligned to the expectations of most users and is already desired in most cases."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Find apotentially affected CASE WHEN statements**\n",
    "\n",
    "This script can be used to find potentially affected CASE WHEN queries in your EMS team. It can search Analyses in Process Analytics, Analyses in Studio, as well as Knowledge Models for those CASE WHEN queries."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## STEP 1: Login\n",
    "\n",
    "\n",
    "The following snippet connects to Celonis by using the application key of the ML workbench. Please make sure that this application key has at least the following permissions.\n",
    "\n",
    "\n",
    "### Required Permissions\n",
    "#### For check in Process Analytics\n",
    "In order to check for potentially affected CASE WHEN statements in Process Analytics, assign the following permissions to the ML workbench application key under `Team Settings > Permissions > Process Analytics Permissions`:\n",
    "   - USE ALL ANALYSES\n",
    "   - EDIT ALL ANALYSES\n",
    "   \n",
    "#### For check in Studio Assets\n",
    "In order to check for potentially affected CASE WHEN statements in Studio Assets (Analyses and Knowledge Models), assign the following permissions to the ML workbench application key *for every* Studio Space that should be checked under the `<Three-dot menue of the Space> > Permissions`:\n",
    "   - USE ALL PACKAGES\n",
    "   - EDIT ALL PACKAGES\n",
    "\n",
    "In both cases, the `EDIT ALL` permission is required to access the unpublished versions of Analyses. This script is read-only and will not modify any EMS application."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from pycelonis import get_celonis\n",
    "from pycelonis.celonis_api import api_utils\n",
    "from pycelonis.celonis_api.studio.analysis import Analysis as StudioAnalysis\n",
    "from pycelonis.celonis_api.studio.knowledge_model import KnowledgeModel\n",
    "import pycelonis.celonis_api.studio.node_factory as node_factory\n",
    "import re\n",
    "import pandas as pd\n",
    "\n",
    "\n",
    "celonis = get_celonis(read_only=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## STEP 2: Helper functions\n",
    "\n",
    "These are helper functions. Simply execute the cell below. It does not do anything (yet). Please do not modify the code."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class ProcessAnalyticsResult:\n",
    "    def __init__(self, workspace, analysis, sheet, component, location, query):\n",
    "        self.workspace=workspace\n",
    "        self.analysis=analysis\n",
    "        self.sheet=sheet\n",
    "        self.component=component\n",
    "        self.location=location\n",
    "        self.query=query   \n",
    "        \n",
    "class StudioAnalysisResult:\n",
    "    def __init__(self, space, package, analysis, sheet, component, location, query):\n",
    "        self.space=space\n",
    "        self.package=package\n",
    "        self.analysis=analysis\n",
    "        self.sheet=sheet\n",
    "        self.component=component\n",
    "        self.location=location\n",
    "        self.query=query\n",
    "        \n",
    "class KnowledgeModelResult:\n",
    "    def __init__(self, space, package, knowledge_model, location, id, query):\n",
    "        self.space=space\n",
    "        self.package=package\n",
    "        self.knowledge_model=knowledge_model\n",
    "        self.location=location\n",
    "        self.id=id\n",
    "        self.query=query\n",
    "        \n",
    "        \n",
    "def result_to_csv(result: list, file_name:str):\n",
    "    df =  pd.DataFrame([r.__dict__ for r in result ])\n",
    "    df[\"query\"] = df[\"query\"].apply( lambda x: x.replace(\"\\n\", \"    \") )\n",
    "    df.to_csv(path_or_buf=file_name, sep=\",\")\n",
    "    print(\"Wrote potentially affected CASE WHEN statements to file {}\".format(file_name))\n",
    "    \n",
    "    \n",
    "def has_case_when(pql: str):\n",
    "    return re.search(r\"\\s+ELSE(?!\\s+NULL\\s+END)\", str(pql), re.IGNORECASE)\n",
    "\n",
    "\n",
    "def create_analysis_result(analysis, sheet, component, location, query, space=None, package=None):\n",
    "    if package:\n",
    "        # studio analysis\n",
    "        return StudioAnalysisResult(space.name, package.name, analysis.name, sheet, component,location, query)\n",
    "    else:\n",
    "        # process analytics\n",
    "        return ProcessAnalyticsResult(analysis.workspace.name, analysis.name, sheet, component,location, query)\n",
    "\n",
    "\n",
    "def find_in_analysis(analysis, space=None, package=None):    \n",
    "    affected_case_whens=[]\n",
    "    for saved_formula in analysis.saved_formulas:\n",
    "        if (has_case_when(saved_formula.data[\"template\"])):\n",
    "            affected_case_whens.append(create_analysis_result(analysis, \"\", \"\", \"Saved formula {}\".format(saved_formula.data[\"name\"]), saved_formula.data[\"template\"], space=space, package=package))\n",
    "    \n",
    "    document=analysis.draft\n",
    "    \n",
    "    if document is None:\n",
    "        return\n",
    "    if document.data[\"document\"] is None:\n",
    "        return\n",
    "    \n",
    "    if document.pql_filter is not None:\n",
    "        if has_case_when(document.pql_filter):\n",
    "            affected_case_whens.append(create_analysis_result(analysis, \"\", \"\", \"Analysis Load Script\", document.pql_filter, space=space, package=package))\n",
    "            \n",
    "    for variable in document.variables:\n",
    "        if \"value\" in variable and has_case_when(variable[\"value\"]):\n",
    "            affected_case_whens.append(create_analysis_result(analysis, \"\", \"\", \"Variable {}\".format(variable[\"name\"]), variable[\"value\"], space=space, package=package))\n",
    "    \n",
    "    for sheet in document.sheets:\n",
    "        if sheet.pql_filter is not None:\n",
    "            if has_case_when(sheet.pql_filter):\n",
    "                affected_case_whens.append(create_analysis_result(analysis, sheet.name, \"\", \"Sheet Filter\", sheet.pql_filter, space=space, package=package))\n",
    "        \n",
    "        for component in sheet.components:\n",
    "            if component is not None:\n",
    "                component_name = \"Component {}\".format(component.name if component.name else \"<unnamed>\")\n",
    "                if component.pql_filter is not None:\n",
    "                    if has_case_when(component.pql_filter):\n",
    "                        affected_case_whens.append(create_analysis_result(analysis, sheet.name, component_name, \"Component Filter\", component.pql_filter, space=space, package=package))\n",
    "                \n",
    "                try:\n",
    "                    if component.pql_columns is not None:\n",
    "                        pql_columns = component.pql_columns\n",
    "                        pql_columns = pql_columns if isinstance(pql_columns, list) else [pql_columns]\n",
    "                        col_id = 0\n",
    "                        for pql_column in pql_columns:\n",
    "                            col_id = col_id +1\n",
    "                            column_name = \"Column {} ({})\".format(col_id, pql_column.name if pql_column.name else \"<unnamed>\")\n",
    "                            if hasattr(pql_column, \"query\"):\n",
    "                                if has_case_when(pql_column.query):\n",
    "                                    affected_case_whens.append(create_analysis_result(analysis, sheet.name, component_name, column_name, pql_column.query, space=space, package=package))\n",
    "                except:\n",
    "                    pass\n",
    "                                \n",
    "    return affected_case_whens\n",
    "\n",
    "\n",
    "def find_in_process_analytics(analysis):\n",
    "    print(\"Checking Analysis {} in Workspace {}.\".format(analysis.name, analysis.workspace.name))\n",
    "    return find_in_analysis(analysis)\n",
    "    \n",
    "def find_in_studio_analysis(analysis, space, package):\n",
    "    print(\"Checking Studio Analysis {} in Package {}/{}.\".format(analysis.name, space.name, package.name))\n",
    "    return find_in_analysis(analysis, space, package)\n",
    "    \n",
    "def find_in_knowledge_model(knowledge_model, space, package):\n",
    "    print(\"Checking Studio Knowledge Model {} in Package {}/{}.\".format(knowledge_model.name, space.name, package.name))\n",
    "\n",
    "    affected_case_whens=[]\n",
    "    if knowledge_model is not None:\n",
    "        for kpi in knowledge_model.raw_kpis:\n",
    "            if \"pql\" in kpi and has_case_when(kpi[\"pql\"]):\n",
    "                affected_case_whens.append(KnowledgeModelResult(space.name, package.name, knowledge_model.name, \"KPI\", kpi[\"id\"], kpi[\"pql\"]))\n",
    "        for record in knowledge_model.raw_records:\n",
    "            if \"pql\" in record and has_case_when(record[\"pql\"]):\n",
    "                affected_case_whens.append(KnowledgeModelResult(space.name, package.name, knowledge_model.name, \"Record\", record[\"id\"], record[\"pql\"]))\n",
    "        for filter in knowledge_model.raw_filters:\n",
    "            if \"pql\" in filter and has_case_when(filter[\"pql\"]):\n",
    "                affected_case_whens.append(KnowledgeModelResult(space.name, package.name, knowledge_model.name, \"Filter\", filter[\"id\"], filter[\"pql\"]))\n",
    "    return affected_case_whens\n",
    "\n",
    "\n",
    "def new_knowledge_model(package, celonis, id):\n",
    "    try:\n",
    "        return KnowledgeModel(package, celonis, id)\n",
    "    except:\n",
    "        return None\n",
    "    \n",
    "        \n",
    "def get_studio_assets_in_package(package, asset_type):\n",
    "    r = celonis.api_request(package.package_nodes_url, threaded=True)\n",
    "        \n",
    "    return [x for x in [\n",
    "            StudioAnalysis(package, celonis, data[\"id\"]) if data[\"assetType\"] == \"ANALYSIS\" else new_knowledge_model(package, celonis, data[\"id\"]) if data[\"assetType\"] == \"SEMANTIC_MODEL\" else None\n",
    "            for data in r\n",
    "            if data[\"nodeType\"] == 'ASSET' and data[\"assetType\"] == asset_type\n",
    "        ] if x is not None]\n",
    "\n",
    "\n",
    "def get_studio_knowledge_model_in_package(package, asset_type):\n",
    "    r = celonis.api_request(package.package_nodes_url, threaded=True)\n",
    "        \n",
    "    return [\n",
    "            KnowledgeModel(package, celonis, data[\"id\"])\n",
    "            for data in r\n",
    "            if data[\"nodeType\"] == 'ASSET' and data[\"assetType\"] == \"SEMANTIC_MODEL\"\n",
    "        ]\n",
    "\n",
    "\n",
    "## PROCESS ANALYTICS\n",
    "def check_process_analytics_analyses():\n",
    "    analyses = celonis.analyses\n",
    "    result = []\n",
    "    for analysis in analyses:\n",
    "        result.extend(find_in_process_analytics(analysis))\n",
    "    if result:\n",
    "        print(\"{} potentially affected CASE WHEN statements found in Process Analytics.\".format(len(result)))\n",
    "        result_to_csv(result, \"process_analytics.csv\")\n",
    "    else:\n",
    "        print(\"No potentially affected CASE WHEN statements found in Process Analytics.\")\n",
    "        \n",
    "\n",
    "## STUDIO ANALYSES\n",
    "def check_studio_analyses():\n",
    "    result = []\n",
    "    spaces = celonis.spaces\n",
    "    for space in spaces:\n",
    "        for package in space.packages:\n",
    "            for analysis in get_studio_assets_in_package(package, \"ANALYSIS\"):\n",
    "                result.extend(find_in_studio_analysis(analysis, space, package))\n",
    "    if result:\n",
    "        print(\"{} potentially affected CASE WHEN statements found in Studio Analyses.\".format(len(result)))\n",
    "        result_to_csv(result, \"studio_analyses.csv\")\n",
    "    else:\n",
    "        print(\"No potentially affected CASE WHEN statements found in Studio Analyses.\")\n",
    "\n",
    "\n",
    "## KNOWLEDGE MODEL\n",
    "def check_knowledge_models():\n",
    "    result=[]\n",
    "    spaces = celonis.spaces\n",
    "    for space in spaces:\n",
    "        for package in space.packages:\n",
    "            for knowledge_model in get_studio_assets_in_package(package, \"SEMANTIC_MODEL\"):\n",
    "                result.extend(find_in_knowledge_model(knowledge_model, space, package))\n",
    "    if result:\n",
    "        print(\"{} potentially affected CASE WHEN statements found in Studio Knowledge Models.\".format(len(result)))\n",
    "        result_to_csv(result, \"studio_knowledge_models.csv\")\n",
    "    else:\n",
    "        print(\"No potentially affected CASE WHEN statements found in Studio Knowledge Models.\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## STEP 3: Execute checks"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### STEP 3.1: Check Process Analytics Analyses\n",
    "\n",
    "Execute the following cell if you want to find potentially affected CASE WHEN statements inside Process Analytics Analyses. It will generate a file called `process_analytics.csv` with the found CASE WHEN statements and their exact locations."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "check_process_analytics_analyses()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### STEP 3.2: Check Studio Analyses\n",
    "\n",
    "Execute the following cell if you want to find potentially affected CASE WHEN statements inside Studio Analyses. It will generate a file called `studio_analyses.csv` with the found CASE WHEN statements and their exact locations."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "check_studio_analyses()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### STEP 3.1: Check Knowledge Models\n",
    "\n",
    "Execute the following cell if you want to find potentially affected CASE WHEN statements inside Studio Knowledge Models. It will generate a file called `studio_knowledge_models.csv` with the found CASE WHEN statements and their exact locations."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "check_knowledge_models()"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "formats": "ipynb,py:percent"
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
